# -*- coding: utf-8 -*-
########################### Empirical Application ############################
# Paper: "Davezies, D'Haultfoeuille, Mugnier (2022)"
# File: 3_generalized_logit_estimation.py
# Task: This code outputs Table 1 in Davezies, D'Haultfoeuille, Mugnier (2022)
#
# Author: Martin Mugnier, martin.mugnier@ensae.fr
# Date: October 11, 2022
##############################################################################

# set path to replication folder 	 
path = 'W:/Documents/SIMULATIONS CLUSTER/application_extension_logit_summer_2022'
import os, warnings
os.chdir(path)

# import packages & estimators
import pandas as pd
import numpy as np
from tqdm import tqdm
from PanelPreProcess import PanelPreProcess
from permGLogit import permGLogit

PanelPreProcess = PanelPreProcess()

# import the data
df_nar = pd.read_csv('narrow_data.csv')
df_exp = pd.read_csv('expanded_data.csv')

dev_nar = df_nar['dev'] == 1
dev_exp = df_exp['dev'] == 1

XX_nar = df_nar[['code', 'year', 'ddef3_n', 'ddef1', 'gdppc_gr2', 'nd', 'maj']]
yy_nar = df_nar[['reelect2']]
XX_exp = df_exp[['code', 'year', 'ddef3_n', 'ddef1', 'gdppc_gr3', 'nd', 'maj']]
yy_exp = df_exp[['reelect3']]

# keep developped countries only
X_nar = XX_nar[dev_nar]
y_nar = yy_nar[dev_nar]
X_exp = XX_exp[dev_exp]
y_exp = yy_exp[dev_exp]

# reshape to N x T x K Numpy array
X_nar, y_nar = PanelPreProcess.sort_panel(np.array(X_nar), np.array(y_nar), 
                                          0, 1)
X_nar, y_nar, _, _ = PanelPreProcess.reshape_MaR(X_nar, y_nar, 0, 1)
X_exp, y_exp = PanelPreProcess.sort_panel(np.array(X_exp), np.array(y_exp), 0,
                                          1)
X_exp, y_exp, _, _ = PanelPreProcess.reshape_MaR(X_exp, y_exp, 0, 1)
n = X_exp.shape[0]

# print the summary statistics given in the text of Section 4 and the "Notes" 
# section of Table 1.
print('Min number of observations per country in narrow sample: ', np.min(
    np.sum(~np.isnan(y_nar), axis=1)))
print('Min number of observations per country in expanded sample: ', np.min(
    np.sum(~np.isnan(y_exp), axis=1)))
print('Max number of observations per country in narrow sample: ', np.max(
    np.sum(~np.isnan(y_nar), axis=1)))
print('Max number of observations per country in expanded sample: ', np.max(
    np.sum(~np.isnan(y_exp), axis=1)))
print('Mean number of observations per country in narrow sample: ', np.round(
        np.mean(np.sum(~np.isnan(y_nar), axis=1)), 1))
print('Mean number of observations per country in narrow sample: ', np.round(
        np.mean(np.sum(~np.isnan(y_exp), axis=1)), 1))

# import Brender and Drazen (2008)'s estimates and asymptotic variance / n
coeffs_narrow = np.array(pd.read_excel('coeffs_narrow.xlsx', header=None))
coeffs_expanded = np.array(pd.read_excel('coeffs_expanded.xlsx', header=None))
LOGIT_avar_spe_1_narrow = np.array(pd.read_excel(
    'LOGIT_avar_spe_1_narrow.xlsx', header=None))
LOGITFE_avar_spe_2_narrow = np.array(pd.read_excel(
    'LOGITFE_avar_spe_2_narrow.xlsx', header=None))
LOGIT_avar_spe_5_expanded = np.array(pd.read_excel(
    'LOGIT_avar_spe_5_expanded.xlsx', header=None))
LOGITFE_avar_spe_6_expanded = np.array(pd.read_excel(
    'LOGITFE_avar_spe_6_expanded.xlsx', header=None))

# create and fill Table 1 
tab = 9999*np.zeros((18, 6))

# narrow sample 
# sign of BALCH_ey LOGIT (1 if >0, -1 else)
if coeffs_narrow[0, 1]>0:
    tab[0, 0] = 1
else:
    tab[0, 0] = -1
# sign of BALCH_ey FE LOGIT (1 if >0, -1 else <=0)
if coeffs_narrow[1, 1]>0: 
    tab[0, 1] = 1
else:
    tab[0, 1] = -1
# ratios and t-statistics unde the null 
for idx, j in enumerate([0, 2, 3, 4]):
    tab[1+2*idx, 0] = coeffs_narrow[0, j] / coeffs_narrow[0, 1]  
    vcov = LOGIT_avar_spe_1_narrow[j, j] 
    tab[1+2*idx+1, 0] = np.abs(coeffs_narrow[0, j]) / np.sqrt(vcov)
    
    tab[1+2*idx, 1] = coeffs_narrow[1, j] / coeffs_narrow[1, 1]
    vcov = LOGITFE_avar_spe_2_narrow[j, j] 
    tab[1+2*idx+1, 1] = np.abs(coeffs_narrow[1, j]) / np.sqrt(vcov)

# expanded sample 
# sign of BALCH_ey LOGIT (1 for >0, -1 for <=0)
if coeffs_expanded[0, 1]>0:
    tab[9, 0] = 1
else:
    tab[9, 0] = -1
# sign of BALCH_ey FE LOGIT (1 for >0, -1 for <=0)
if coeffs_expanded[1, 1]>0: 
    tab[9, 1] = 1
else:
    tab[9, 1] = -1
# ratios and t-statistics unde the null 
for idx, j in enumerate([0, 2, 3, 4]):
    tab[10+2*idx, 0] = coeffs_expanded[0, j] / coeffs_expanded[0, 1]  
    vcov = LOGIT_avar_spe_5_expanded[j, j] 
    tab[10+2*idx+1, 0] = np.abs(coeffs_expanded[0, j]) / np.sqrt(vcov)
    
    tab[10+2*idx, 1] = coeffs_expanded[1, j] / coeffs_expanded[1, 1]  
    vcov = LOGITFE_avar_spe_6_expanded[j, j] 
    tab[10+2*idx+1, 1] = np.abs(coeffs_expanded[1, j]) / np.sqrt(vcov)
    
# GMM estimation of FE generalized Logit for different values of lambda_2
lbda1 = 1
lbda2_grid = [1.2, 1.4, 1.6, 1.8] 
K = len(LOGITFE_avar_spe_2_narrow) # number of covariates
weighting_mat = np.eye(3*K) # GMM weigthing matrix
np.random.seed(2022) # for replicability
n_init = 200 # number of random initializers
init_narrow = coeffs_narrow[1,:-1] # initialized around CMLE
init_exp = coeffs_expanded[1,:-1] # initialized arround CMLE
warnings.filterwarnings('ignore') # suppress warnings (overflows in exponentials
                                    # may happen during optimization)
for idxl, lbda2 in enumerate(lbda2_grid):   
    lbda = [lbda1, lbda2]
    # instantiate the model in each sample
    model_nar = permGLogit(X_nar, y_nar, lbda, weighting_mat, demeaned=True)
    model_exp = permGLogit(X_exp, y_exp, lbda, weighting_mat, demeaned=True)
    beta_res = np.zeros((n_init, K, 2))
    GMM_crit = np.zeros((n_init, 2))
    for b in tqdm(range(n_init)):
        # estimate the model for given initializers in each sample
        beta_res[b,:, 0], GMM_crit[b, 0] = model_nar.fit(init_narrow
                + np.random.uniform(-0.2, 0.2, size=K))
        beta_res[b,:, 1], GMM_crit[b, 1] = model_exp.fit(init_exp
                + np.random.uniform(-0.2, 0.2, size=K))
    # keep estimates with lowest GMM criterion value
    glfe_nar = beta_res[np.nanargmin(GMM_crit[:,0]),:,0]
    glfe_exp = beta_res[np.nanargmin(GMM_crit[:,1]),:,1]
    vcov_nar = model_nar.avar(glfe_nar) / n
    vcov_exp = model_exp.avar(glfe_exp) / n

    # narrow sample 
    # sign of BALCH_ey FE generalized LOGIT (1 if >0, -1 else)
    if glfe_nar[1]>0:
        tab[0, 2+idxl] = 1
    else:
        tab[0, 2+idxl] = -1
    # ratios and t-statistics unde the null 
    for idx, j in enumerate([0, 2, 3, 4]):
        tab[1+2*idx, 2+idxl] = glfe_nar[j] / glfe_nar[1]  
        tab[1+2*idx+1, 2+idxl] = np.abs(glfe_nar[j]) / np.sqrt(vcov_nar[j, j])
    
    # expanded sample 
    # sign of BALCH_ey FE generalized LOGIT (1 if >0, -1 else)
    if glfe_exp[1]>0:
        tab[9, 2+idxl] = 1
    else:
        tab[9, 2+idxl] = -1
    # ratios and t-statistics unde the null 
    for idx, j in enumerate([0, 2, 3, 4]):
        tab[10+2*idx, 2+idxl] = glfe_exp[j] / glfe_exp[1] 
        tab[10+2*idx+1, 2+idxl] = np.abs(glfe_exp[j])/np.sqrt(
            vcov_exp[j, j])

# export Table 1 as a .tex file 
res = pd.DataFrame(np.round(tab,2))
print("Narrow sample: BALCH_eye coeff always positive?", np.sum(tab[0])==6)
res.iloc[0] = [">0", ">0", ">0", ">0", ">0", ">0"]
print("Expanded sample: BALCH_eye coeff always positive?", np.sum(tab[9])==6)
res.iloc[9] = [">0", ">0", ">0", ">0", ">0", ">0"]
res.to_latex('table_1_analytical_tstats_'+str(n_init)
                +'_initialization_points.tex') 
