# -*- coding: utf-8 -*-
# ensure "normal" division
from __future__ import division

import numpy as np
from itertools import permutations
from scipy.optimize import minimize

# useful functions
is_sorted = lambda a: np.all(a[:-1]<=a[1:])

def multi_outer(A,B):
    
    """
    multi_outer:
        This function computes rowwise outer products between matrices A and B.
        
    INPUTS:
    ------
    `A'       : N x K (Numpy) array;
    `B'       : N x K (Numpy) array.
    
    OUTPUT:
    ------
    `C''       : N x K x K (Numpy) array C with C[i,:,:] = A[i,:]*B[i,:].T.
    """
    
    return A[:,:,None]*B[:,None,:]

class permGLogit():
    
    """
    permGLogit:
        This class implements the permutation GMM estimator for estimating 
        fixed effects binary choice models with generalized logistic errors 
        as described in Davezies, D'Haultfoeuille and Mugnier (2022, 
        Working Paper).
        -- ``Fixed Effects Binary Choice Models with Three or More Periods''. 
        The implementation is as described in the paper. Notation attempts to 
        follow that used in the paper. In particular: 
        `N'  : number of individuals;
        `T'  : number of time periods (fill missing entries with Nan);
        `K'  : number of observed covariates.
        
    AUTHOR:         Martin Mugnier, martin.mugnier@ensae.fr, August 2022
    ------
    
    INPUTS:
    ------
    `X'            : N x T x K (Numpy) array of exogeneous variables;
    `Y'            : N x T (Numpy) array of scalar endogeneous variables;
    `lbda'         : 2 x 1 (Numpy) array of parameters of the generalized
                    logistic distribution.

    FUNCTIONS:
    ---------
    `mu'
    `grad_mu'
    `hess_mu'
    `instruments'
    `moment'
    `jac_moment'
    `hess_moment'
    `GMM_crit''
    `GMM_crit_jac'
    `GMM_crit_hess'
    `fit'
    `newton_raphson'
    `avar'
    
    """
    
    def __init__(self, X, Y, lbda, weight_mat=np.eye(6), norm=False, 
                 demeaned=False):
        self.X = X
        self.Y = Y
        self.lbda = lbda
        self.weight_mat = weight_mat
        self.norm = norm
        self.demeaned = demeaned
        self.N, self.T, self.K = X.shape
        self.perms = [] 
         
        for i in range(self.N):
            # triads of time periods with no missing outcome
            a = np.array([elem for elem in 
                          permutations(np.where(~np.isnan(self.Y[i]))[0], 3)
                         if is_sorted(elem)])
            # keep only if the triad outcomes sum to one 
            self.perms.append(np.array([elem for elem in a if 
                                        np.sum(self.Y[i,elem])==1]))
        # mean number of triads per individual
        self.mean_perms = np.mean([len(elem) for elem in self.perms])
        # units effectively used in estimation (i.e., with at least one triad)
        self.informative_units = np.arange(self.N)[np.array([int(len(
                self.perms[i])) for i in range(self.N)])>0]
    
    def mu(self, x, y, beta):
        
        """
        mu:
            compute the moment datum.
            
        INPUTS:
        ------
        `x'        : 3 x K (Numpy) array of individual-specific covariates;
        `y'        : 3 x 1 (Numpy) array of individual-specific outcomes;
        `beta'     : K x 1 (Numpy) array of slope coefficients.
        
        OUTPUT:
        ------
        mu         : scalar (for a given triad).
        
        """
        
        idx = x.dot(beta)
        mu = np.zeros(3)
        mu[0] = np.exp(self.lbda[0]*idx[1] + self.lbda[1]*idx[2]) \
                    - np.exp(self.lbda[1]*idx[1] + self.lbda[0]*idx[2])
        mu[1] = np.exp(self.lbda[0]*idx[2] + self.lbda[1]*idx[0]) \
                    - np.exp(self.lbda[1]*idx[2] + self.lbda[0]*idx[0])
        mu[2] = np.exp(self.lbda[0]*idx[0] + self.lbda[1]*idx[1]) \
                    - np.exp(self.lbda[1]*idx[0] + self.lbda[0]*idx[1])
        res = np.sum(mu*y)
        if self.norm:
            res /= np.linalg.norm(mu)
        return  res

    def grad_mu(self, x, y, beta):
        
        """
        grad_mu:
            compute gradient of the moment datum.
            
        INPUTS:
        ------
        `x'        : 3 x K (Numpy) array of individual-specific covariates;
        `y'        : 3 x 1 (Numpy) array of individual-specific outcomes;
        `beta'     : K x 1 (Numpy) array of slope coefficients.
        
        OUTPUT:
        ------
        grad_mu    : K x 1 (Numpy) array (for a given triad).
        
        """
        
        idx = x.dot(beta)
        gmu = np.zeros((3, self.K))
        gmu[0] = (self.lbda[0]*x[1] + self.lbda[1]*x[2]) *  \
                    np.exp(self.lbda[0]*idx[1] + self.lbda[1]*idx[2]) \
                    - (self.lbda[1]*x[1] + self.lbda[0]*x[2]) * \
                    np.exp(self.lbda[1]*idx[1] + self.lbda[0]*idx[2])
        gmu[1] = (self.lbda[0]*x[2] + self.lbda[1]*x[0]) * \
                    np.exp(self.lbda[0]*idx[2] + self.lbda[1]*idx[0]) \
                    - (self.lbda[1]*x[2] + self.lbda[0]*x[0]) * \
                    np.exp(self.lbda[1]*idx[2] + self.lbda[0]*idx[0])
        gmu[2] = (self.lbda[0]*x[0] + self.lbda[1]*x[1]) * \
                    np.exp(self.lbda[0]*idx[0] + self.lbda[1]*idx[1])  \
                    - (self.lbda[1]*x[0] + self.lbda[0]*x[1]) * \
                    np.exp(self.lbda[1]*idx[0] + self.lbda[0]*idx[1])
        if self.norm:
            mu = np.zeros(3)
            mu[0] = np.exp(self.lbda[0]*idx[1] + self.lbda[1]*idx[2]) \
                    - np.exp(self.lbda[1]*idx[1] + self.lbda[0]*idx[2])
            mu[1] = np.exp(self.lbda[0]*idx[2] + self.lbda[1]*idx[0]) \
                    - np.exp(self.lbda[1]*idx[2] + self.lbda[0]*idx[0])
            mu[2] = np.exp(self.lbda[0]*idx[0] + self.lbda[1]*idx[1]) \
                    - np.exp(self.lbda[1]*idx[0] + self.lbda[0]*idx[1])
            normalization = np.linalg.norm(mu)  
            res = (np.sum(gmu*y[:,None], axis=0)*normalization
                - np.transpose(gmu).dot(mu)*np.sum(mu*y)/normalization) \
                / (normalization**2)
        else:
             res = np.sum(gmu*y[:,None], axis=0) 
        return res
    
    def hess_mu(self, x, y, beta):
        
        """
        hess_mu:
            compute Hessian of the moment datum.
            
        INPUTS:
        ------
        `x'        : 3 x K (Numpy) array of individual-specific covariates;
        `y'        : 3 x 1 (Numpy) array of individual-specific outcomes;
        `beta'     : K x 1 (Numpy) array of slope coefficients.
        
        OUTPUT:
        ------
        hess_mu    : K x K (Numpy) array.
        
        """
        
        idx = x.dot(beta)
        hmu = np.zeros((3, self.K, self.K))
        hmu[0] = (self.lbda[0]*x[1] + self.lbda[1]*x[2])[:,None] *  \
             (self.lbda[0]*x[1] + self.lbda[1]*x[2])[None:] * \
            np.exp(self.lbda[0]*idx[1] + self.lbda[1]*idx[2]) \
            - (self.lbda[1]*x[1] + self.lbda[0]*x[2])[:,None] * \
            (self.lbda[1]*x[1] + self.lbda[0]*x[2])[None,:] * \
            np.exp(self.lbda[1]*idx[1] + self.lbda[0]*idx[2])
        hmu[1] = (self.lbda[0]*x[2] + self.lbda[1]*x[0])[:,None] * \
                (self.lbda[0]*x[2] + self.lbda[1]*x[0])[None,:] * \
                    np.exp(self.lbda[0]*idx[2] + self.lbda[1]*idx[0]) \
                    - (self.lbda[1]*x[2] + self.lbda[0]*x[0])[:,None] * \
                    (self.lbda[1]*x[2] + self.lbda[0]*x[0])[None,:] * \
                    np.exp(self.lbda[1]*idx[2] + self.lbda[0]*idx[0])
        hmu[2] = (self.lbda[0]*x[0] + self.lbda[1]*x[1])[:,None] * \
                (self.lbda[0]*x[0] + self.lbda[1]*x[1])[None,:] * \
                    np.exp(self.lbda[0]*idx[0] + self.lbda[1]*idx[1])  \
                    - (self.lbda[1]*x[0] + self.lbda[0]*x[1])[:,None] * \
                    (self.lbda[1]*x[0] + self.lbda[0]*x[1])[None,:] * \
                    np.exp(self.lbda[1]*idx[0] + self.lbda[0]*idx[1])
        if self.norm:
            mu = np.zeros(3)
            mu[0] = np.exp(self.lbda[0]*idx[1] + self.lbda[1]*idx[2]) \
                        - np.exp(self.lbda[1]*idx[1] + self.lbda[0]*idx[2])
            mu[1] = np.exp(self.lbda[0]*idx[2] + self.lbda[1]*idx[0]) \
                        - np.exp(self.lbda[1]*idx[2] + self.lbda[0]*idx[0])
            mu[2] = np.exp(self.lbda[0]*idx[0] + self.lbda[1]*idx[1]) \
                        - np.exp(self.lbda[1]*idx[0] + self.lbda[0]*idx[1])
            gmu = np.zeros((3, self.K))
            gmu[0] = (self.lbda[0]*x[1] + self.lbda[1]*x[2]) *  \
                        np.exp(self.lbda[0]*idx[1] + self.lbda[1]*idx[2]) \
                        - (self.lbda[1]*x[1] + self.lbda[0]*x[2]) * \
                        np.exp(self.lbda[1]*idx[1] + self.lbda[0]*idx[2])
            gmu[1] = (self.lbda[0]*x[2] + self.lbda[1]*x[0]) * \
                        np.exp(self.lbda[0]*idx[2] + self.lbda[1]*idx[0]) \
                        - (self.lbda[1]*x[2] + self.lbda[0]*x[0]) * \
                        np.exp(self.lbda[1]*idx[2] + self.lbda[0]*idx[0])
            gmu[2] = (self.lbda[0]*x[0] + self.lbda[1]*x[1]) * \
                        np.exp(self.lbda[0]*idx[0] + self.lbda[1]*idx[1])  \
                        - (self.lbda[1]*x[0] + self.lbda[0]*x[1]) * \
                        np.exp(self.lbda[1]*idx[0] + self.lbda[0]*idx[1])
            normalization = np.linalg.norm(mu)
            res = ((np.sum(hmu*y[:,None,None], axis=0) * normalization**2 + 
                2*np.sum(gmu*y[:,None], axis=0)[:,None]*np.transpose(gmu).dot(
                    mu)[None,:]
               - (np.sum(hmu*mu[:,None,None] + gmu[:,:,None]*gmu[:,None,:], 
                  axis=0)*np.sum(mu*y) + np.sum(gmu*y[:,None], axis=0)[:,None]*
                      np.transpose(gmu).dot(mu)[None,:])) * normalization**3 -
                      (3 *np.transpose(gmu).dot(mu)*normalization**2)[:,None] *
                      (np.sum(gmu*y[:,None], axis=0) * normalization**2
                       -  np.transpose(gmu).dot(mu) * np.sum(mu*y))[None,:]) \
                        / (normalization**6 + 1e-10)
        else:
            res = np.sum(hmu*y[:,None,None], axis=0)
        return res
    
    def instruments(self, x):
        
        """
        instruments:
            compute the instruments (all covariates at all triad periods).
            
        INPUTS:
        ------
        `x'        : 3 x K (Numpy) array of individual-specific covariates.
        
        OUTPUT:
        ------
        res        : 3*K x 1 (Numpy) array.
        
        """
        
        res = x.reshape(3*self.K)
        return res
    
    def moment(self, beta, mean=1):
        
        """
        moment:
            compute the empirical moment function.
            
        INPUTS:
        ------
        `beta'  : K x 1 (Numpy) array of individual-specific covariates.
        
        OUTPUT:
        ------
        m       : 3*K x 1 (Numpy) array if mean==1, N x 3*K (Numpy) array else.
        
        """
        
        m = np.zeros((self.N, 3*self.K))
        if self.demeaned:
            for i in self.informative_units:
                for elem in self.perms[i]:
                    m[i] += self.mu(self.X[i, elem,:], self.Y[i, elem], beta) \
                    * self.instruments(self.X[i,elem,:]-np.mean(
                        self.X[i,elem,:], axis=0))
        else:
            for i in self.informative_units:
                    for elem in self.perms[i]:
                        m[i] += self.mu(self.X[i, elem,:], self.Y[i, elem], 
                        beta) * self.instruments(self.X[i,elem,:])
        if mean:
            m = np.mean(m, axis=0)
        return m
    
    def jac_moment(self, beta):
        
        """
        jac_moment:
            compute the jacobian matrix of the empirical moment function.
            
        INPUTS:
        ------
        `beta'     : K x 1 (Numpy) array of individual-specific covariates.
        
        OUTPUT:
        ------
        j          : 3*K x K (Numpy) array.
        
        """
        
        j = np.zeros((3*self.K, self.K))
        if self.demeaned:
            for i in self.informative_units:
                for elem in self.perms[i]:
                    j += self.grad_mu(self.X[i,elem], self.Y[i,elem],
                        beta)[None,:]*self.instruments(self.X[i,elem,:]-
                        np.mean(self.X[i,elem,:], axis=0))[:,None]
        else:
            for i in self.informative_units:
                for elem in self.perms[i]:
                    j += self.grad_mu(self.X[i,elem], self.Y[i,elem],
                        beta)[None,:]* self.instruments(
                            self.X[i,elem,:])[:,None] 
        return j / self.N
    
    def hess_moment(self, beta): 
        """
        hess_moment:
            compute Hessian (3-dim) matrix of the empirical moment function.
            
        INPUTS:
        ------
        `beta'     : K x 1 (Numpy) array of individual-specific covariates.
        
        OUTPUT:
        ------
        j          : K x K x 3*K (Numpy) array.
        
        """ 
        h = np.zeros((self.K, self.K, 3*self.K))
        if self.demeaned:
            for i in self.informative_units:
                for elem in self.perms[i]:
                    h += self.instruments(self.X[i,elem,:]-np.mean(
                        self.X[i,elem,:], axis=0))[None,None,:] * \
                      self.hess_mu(self.X[i,elem], self.Y[i,elem], 
                                   beta)[:,:,None]
        else:
            for i in self.informative_units:
                for elem in self.perms[i]:
                    h += self.instruments(self.X[i,elem,:])[None,None,:] * \
                      self.hess_mu(self.X[i,elem], self.Y[i,elem], 
                                   beta)[:,:,None]
        return h / self.N
    
    def GMM_crit(self, beta):
        
        """
        GMM_crit:
            compute the GMM objective function.
            
        INPUTS:
        ------
        `beta'     : K x 1 (Numpy) array of individual-specific covariates.
        
        OUTPUT:
        ------
        GMM_crit   : scalar.
        
        """
        
        return self.moment(beta).dot(self.weight_mat).dot(self.moment(beta))
    
    def GMM_crit_jac(self, beta):
        
        """
        GMM_crit_jac:
            compute the jacobian of the GMM objective function.
            
        INPUTS:
        ------
        `beta'       : K x 1 (Numpy) array of individual-specific covariates.
        
        OUTPUT:
        ------
        GMM_crit_jac : 1 x K (Numpy) array.
        
        """
        
        return 2*self.jac_moment(beta).T.dot(self.weight_mat).dot(
            self.moment(beta))
    
    def GMM_crit_hess(self, beta):
        
        """
        GMM_crit_hess:
            compute the Hessian of the GMM objective function.
            
        INPUTS:
        ------
        `beta'         : K x 1 (Numpy) array of individual-specific covariates.
        
        OUTPUT:
        ------
        GMM_crit_hess  : K x K (Numpy) array.
        
        """
        
        return self.hess_moment(beta).dot(self.weight_mat).dot(
            self.moment(beta)) + self.jac_moment(beta).T.dot(
                self.weight_mat).dot(self.jac_moment(beta)) 

    def fit(self, beta_init):
        
        """
        fit:
            estimate the parameters. 
            
        INPUTS:
        ------
        `beta_init'    : K x 1 (Numpy) array of intial guess.
        
        OUTPUT:
        ------
        res.x          : K x 1 (Numpy) array of estimated parameters;
        res.fun        : scalar objective value.
        
        """
        
        res = minimize(self.GMM_crit, beta_init, jac=self.GMM_crit_jac,
                       method='L-BFGS-B')
        return res.x, res.fun
    
    def newton_raphson(self, beta_init, eps=1e-03, max_iter=100):
        
        """
        newton_raphson:
            Newton-Raphson optimization algorithm. 
            
        INPUTS:
        ------
        `beta_init'    : K x 1 (Numpy) array of intial guess;
        `eps'          : scalar tolerance;
        `max_iter'     : maximum number of iterations.
        
        OUTPUT:
        ------
        res.x          : K x 1 (Numpy) array of estimated parameters;
        res.fun        : scalar objective value.
        
        """
        
        x = [0]
        niter=1
        xtol = eps+1
        beta = beta_init.copy()
        while (niter<=max_iter)*(xtol>=eps):
            beta_new = beta - np.linalg.inv(self.GMM_crit_hess(beta)).dot(
                self.GMM_crit_jac(beta))
            x.append(beta_new)
            xtol = np.linalg.norm(beta-beta_new)
            beta = beta_new
            niter +=1
        return beta, self.GMM_crit(beta)
    
    def avar(self, beta):
        
        """
        avar:
            compute an estimate of the analytic formula of the GMM asymptotic 
            variance-covariance matrix. 
            
        INPUTS:
        ------
        `beta'         : K x 1 (Numpy) array of estimated parameter.
        
        OUTPUT:
        ------
        avar          : K x K (Numpy) array.
        
        """
        
        G = self.jac_moment(beta)
        Sig = np.mean(multi_outer(self.moment(beta, mean=0), self.moment(beta, 
                                  mean=0)), axis=0)
        B =  np.linalg.inv(G.T.dot(self.weight_mat).dot(G))  
        avar = B.dot(G.T.dot(self.weight_mat).dot(Sig).dot(self.weight_mat) \
                     .dot(G)).dot(B)
        return avar
    
